#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# Basic information

import sys
import platform

print "Python:"
print "  Basic version: %s.%s.%s" % (sys.version_info[0],
                                     sys.version_info[1],
                                     sys.version_info[2], )
print "  Full version: " + sys.version.replace('\n', ' ')
print

def c(s):
    return s or "<COULD NOT DETERMINE>"

print "System:"
print "  Type: " + c(platform.system())
print "  Architecture: " + c(platform.architecture()[0])
print "  Machine: " + c(platform.machine())
print "  Platform: " + c(platform.platform())
print "  Processor: " + c(platform.processor())
print

##############################################################################

print "Libraries:"

try:
    import sip
    print "  sip installed."
    print "    version: " + sip.SIP_VERSION_STR
except ImportError:
    print "  sip NOT installed."
print

try:
    import PyQt4.Qt
    print "  PyQt installed."
    print "    Qt version: " + PyQt4.Qt.QT_VERSION_STR
    print "    PyQt version: " + PyQt4.Qt.PYQT_VERSION_STR
except ImportError:
    print "  PyQt NOT installed."
print

try:
    import vtk
    print "  VTK installed."
    print "    VTK short version: " + vtk.vtkVersion().GetVTKVersion()
    print "    VTK full version: " + vtk.vtkVersion().GetVTKSourceVersion()
except ImportError:
    print "  VTK NOT installed."
