# pragma: no testimport
"""This is a target for importing a non-package module.

It is used in the PackageManager's tests to check that the dependency-tracking
name-fixing import override works correctly.
"""
