# pragma: no testimport
name = 'test_import_pkg'
identifier = 'org.vistrails.tests.test_import_pkg'
version = '0.42'

import vistrails.tests.resources.import_targets.test1
import tests.resources.import_targets.test2

import module1
